##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.user.ActivationStatus import ActivationStatus
from cfxdb.gen.user.ActivationType import ActivationType
from cfxdb.gen.user.OrganizationType import OrganizationType
from cfxdb.gen.user.UserRole import UserRole
from cfxdb.user.activation_token import ActivationToken
from cfxdb.user.activation_token_fbs import ActivationTokenFbs
from cfxdb.user.organization import Organization
from cfxdb.user.organization_fbs import OrganizationFbs
from cfxdb.user.user import User
from cfxdb.user.user_fbs import UserFbs
from cfxdb.user.user_mrealm_role import UserMrealmRole
from cfxdb.user.user_mrealm_role_fbs import UserMrealmRoleFbs

__all__ = (
    "User",
    "UserFbs",
    "UserMrealmRole",
    "UserMrealmRoleFbs",
    "ActivationToken",
    "ActivationTokenFbs",
    "Organization",
    "OrganizationFbs",
    "ActivationStatus",
    "ActivationType",
    "OrganizationType",
    "UserRole",
)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit

import txaio

txaio.use_twisted()  # noqa

import pytest
from zlmdb import flatbuffers

from cfxdb.xbr import TokenApproval


def fill_token_approval(token_approval):
    token_approval.tx_hash = os.urandom(32)
    token_approval.block_hash = os.urandom(32)
    token_approval.owner_address = os.urandom(20)
    token_approval.spender_address = os.urandom(20)
    token_approval.value = random.randint(1, 2**256 - 1)


@pytest.fixture(scope="function")
def token_approval():
    _token_approval = TokenApproval()
    fill_token_approval(_token_approval)
    return _token_approval


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_token_approval_roundtrip(token_approval, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = token_approval.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 220

    # create python object from bytes (flatbuffes)
    _token_approval = TokenApproval.cast(data)

    assert _token_approval.tx_hash == token_approval.tx_hash
    assert _token_approval.block_hash == token_approval.block_hash
    assert _token_approval.owner_address == token_approval.owner_address
    assert _token_approval.spender_address == token_approval.spender_address
    assert _token_approval.value == token_approval.value


def test_token_approval_roundtrip_perf(token_approval, builder):
    obj = token_approval.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _token_approval = TokenApproval.cast(data)
        if True:
            assert _token_approval.tx_hash == token_approval.tx_hash
            assert _token_approval.block_hash == token_approval.block_hash
            assert _token_approval.owner_address == token_approval.owner_address
            assert _token_approval.spender_address == token_approval.spender_address
            assert _token_approval.value == token_approval.value

            scratch["value"] += int(_token_approval.tx_hash[0])
            scratch["value"] += int(_token_approval.tx_hash[7])
            scratch["value"] += int(_token_approval.tx_hash[19])

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

from cfxdb.common import ConfigurationElement


class Node(ConfigurationElement):
    """
    Nodes paired with management realms on this master node.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        owner_oid: Optional[UUID] = None,
        pubkey: Optional[str] = None,
        cluster_ip: Optional[str] = None,
        mrealm_oid: Optional[UUID] = None,
        authid: Optional[str] = None,
        authextra: Optional[dict] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of node

        :param label: Optional user label of node

        :param description: Optional user description of node

        :param tags: Optional list of user tags on node

        :param owner_oid: Object owner.

        :param pubkey: The WAMP-cryptosign node public key (32 bytes as HEX encoded string).

        :param cluster_ip: The (private) clustering IP for incoming rlinks to connect to.

        :param mrealm_oid: The management realm the node will be joined on.

        :param authid: The WAMP ``authid`` the node will be authenticated as.

        :param authextra: Optional ``authextra`` the node will be sent to when authenticating.

        :param _unknown: Any unparsed/unprocessed data attributes
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.owner_oid = owner_oid
        self.pubkey = pubkey
        self.cluster_ip = cluster_ip
        self.mrealm_oid = mrealm_oid
        self.authid = authid
        self.authextra = authextra

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.owner_oid != self.owner_oid:
            return False
        if other.pubkey != self.pubkey:
            return False
        if other.cluster_ip != self.cluster_ip:
            return False
        if other.mrealm_oid != self.mrealm_oid:
            return False
        if other.authid != self.authid:
            return False
        if other.authextra != self.authextra:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other management realm to copy data from.
        :type other: instance of :class:`ManagementRealm`
        :return:
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.owner_oid and other.owner_oid) or overwrite:
            self.owner_oid = other.owner_oid
        if (not self.pubkey and other.pubkey) or overwrite:
            self.pubkey = other.pubkey
        if (not self.cluster_ip and other.cluster_ip) or overwrite:
            self.cluster_ip = other.cluster_ip
        if (not self.mrealm_oid and other.mrealm_oid) or overwrite:
            self.mrealm_oid = other.mrealm_oid
        if (not self.authid and other.authid) or overwrite:
            self.authid = other.authid
        if (not self.authextra and other.authextra) or overwrite:
            self.authextra = other.authextra

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "owner_oid": str(self.owner_oid) if self.owner_oid else None,
                "pubkey": self.pubkey,
                "cluster_ip": self.cluster_ip,
                "mrealm_oid": str(self.mrealm_oid) if self.mrealm_oid else None,
                "authid": self.authid,
                "authextra": self.authextra,
            }
        )

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["owner_oid", "pubkey", "cluster_ip", "mrealm_oid", "authid", "authextra"]:
                _unknown[k] = data[k]

        owner_oid = data.get("owner_oid", None)
        assert owner_oid is None or type(owner_oid) == str
        if owner_oid:
            owner_oid = UUID(owner_oid)

        pubkey = data.get("pubkey", None)
        assert pubkey is None or (type(pubkey) == str and len(pubkey) == 64)

        cluster_ip = data.get("cluster_ip", None)
        assert cluster_ip is None or type(cluster_ip)

        mrealm_oid = data.get("mrealm_oid", None)
        assert mrealm_oid is None or type(mrealm_oid) == str
        if mrealm_oid:
            mrealm_oid = UUID(mrealm_oid)

        authid = data.get("authid", None)
        assert authid is None or type(authid) == str

        authextra = data.get("authextra", None)
        assert authextra is None or type(authextra) == dict

        obj = Node(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            owner_oid=owner_oid,
            pubkey=pubkey,
            cluster_ip=cluster_ip,
            mrealm_oid=mrealm_oid,
            authid=authid,
            authextra=authextra,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores information about the series of Ethereum blocks that make up the blockchain.
class Block(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Block()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsBlock(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Block
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Timestamp when record was inserted (Unix epoch time in ns)
    # Block
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Primary key: block number.
    # Block
    def BlockNumber(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Block
    def BlockNumberAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Block
    def BlockNumberLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Block
    def BlockNumberIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Block hash.
    # Block
    def BlockHash(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Block
    def BlockHashAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Block
    def BlockHashLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Block
    def BlockHashIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Number of XBR blockchain log events found in the block.
    # Block
    def CntEvents(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0


def BlockStart(builder):
    builder.StartObject(4)


def Start(builder):
    BlockStart(builder)


def BlockAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    BlockAddTimestamp(builder, timestamp)


def BlockAddBlockNumber(builder, blockNumber):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(blockNumber), 0)


def AddBlockNumber(builder, blockNumber):
    BlockAddBlockNumber(builder, blockNumber)


def BlockStartBlockNumberVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartBlockNumberVector(builder, numElems: int) -> int:
    return BlockStartBlockNumberVector(builder, numElems)


def BlockAddBlockHash(builder, blockHash):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(blockHash), 0)


def AddBlockHash(builder, blockHash):
    BlockAddBlockHash(builder, blockHash)


def BlockStartBlockHashVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartBlockHashVector(builder, numElems: int) -> int:
    return BlockStartBlockHashVector(builder, numElems)


def BlockAddCntEvents(builder, cntEvents):
    builder.PrependUint32Slot(3, cntEvents, 0)


def AddCntEvents(builder, cntEvents):
    BlockAddCntEvents(builder, cntEvents)


def BlockEnd(builder):
    return builder.EndObject()


def End(builder):
    return BlockEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Aggregate usage metering records for management realms. Primary key: (timestamp, mrealm_id).
class MasterNodeUsage(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MasterNodeUsage()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMasterNodeUsage(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # MasterNodeUsage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: Timestamp (end of aggregate interval) recorded in UTC (Unix time in ns).
    # MasterNodeUsage
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Primary key: Management realm ID.
    # MasterNodeUsage
    def MrealmId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MasterNodeUsage
    def MrealmIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MasterNodeUsage
    def MrealmIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MasterNodeUsage
    def MrealmIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Timestamp (start of aggregate interval) recorded in UTC (Unix time in ns).
    # MasterNodeUsage
    def TimestampFrom(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Public key of the CrossbarFX master node that submitted the usage record.
    # MasterNodeUsage
    def Pubkey(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MasterNodeUsage
    def PubkeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MasterNodeUsage
    def PubkeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MasterNodeUsage
    def PubkeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Client IPv4 address (4 bytes) or IPv6 (16 bytes) address of the CrossbarFX master node that submitted the usage record.
    # MasterNodeUsage
    def ClientIpAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MasterNodeUsage
    def ClientIpAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MasterNodeUsage
    def ClientIpAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MasterNodeUsage
    def ClientIpAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Client IP version of the CrossbarFX master node that submitted the usage record.
    # MasterNodeUsage
    def ClientIpVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Client IP port of the CrossbarFX master node that submitted the usage record.
    # MasterNodeUsage
    def ClientIpPort(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Sequence number as sent in the log record by the CF node (started at 0 for CF start and incremented by one on each heartbeat).
    # MasterNodeUsage
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Unix time in ns. This timestamp is from the original received event payload (from CF node clock).
    # MasterNodeUsage
    def Sent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Unix time in ns. Set when this record was processed.
    # MasterNodeUsage
    def Processed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Status of usage metering record.
    # MasterNodeUsage
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Status message for current status.
    # MasterNodeUsage
    def StatusMessage(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Filled after this usage metering record was successfully submitted to the metering service.
    # MasterNodeUsage
    def MeteringId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MasterNodeUsage
    def MeteringIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MasterNodeUsage
    def MeteringIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MasterNodeUsage
    def MeteringIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Number of aggregated records from MworkerLogs
    # MasterNodeUsage
    def Count(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of aggregated and summed MworkerLogs.count records
    # MasterNodeUsage
    def Total(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Nodes by node type
    # MasterNodeUsage
    def Nodes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of node controllers metered.
    # MasterNodeUsage
    def Controllers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of hostmonitor workers metered.
    # MasterNodeUsage
    def Hostmonitors(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of router workers metered.
    # MasterNodeUsage
    def Routers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of container workers metered.
    # MasterNodeUsage
    def Containers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of guest workers metered.
    # MasterNodeUsage
    def Guests(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of proxy workers metered.
    # MasterNodeUsage
    def Proxies(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of XBR market maker workers metered.
    # MasterNodeUsage
    def Marketmakers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP sessions attached to router workers and metered.
    # MasterNodeUsage
    def Sessions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(50))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP CALL messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsCall(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(52))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP YIELD messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsYield(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(54))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP INVOCATION messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsInvocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(56))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP RESULT messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsResult(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(58))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP ERROR messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsError(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(60))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP PUBLISH messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsPublish(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(62))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP PUBLISHED messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsPublished(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(64))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP EVENT messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsEvent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(66))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP REGISTER messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsRegister(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(68))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP REGISTERED messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsRegistered(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(70))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP SUBSCRIBE messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsSubscribe(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(72))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of WAMP SUBSCRIBED messages processed by a router worker and metered.
    # MasterNodeUsage
    def MsgsSubscribed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(74))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def MasterNodeUsageStart(builder):
    builder.StartObject(36)


def Start(builder):
    MasterNodeUsageStart(builder)


def MasterNodeUsageAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    MasterNodeUsageAddTimestamp(builder, timestamp)


def MasterNodeUsageAddMrealmId(builder, mrealmId):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(mrealmId), 0)


def AddMrealmId(builder, mrealmId):
    MasterNodeUsageAddMrealmId(builder, mrealmId)


def MasterNodeUsageStartMrealmIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMrealmIdVector(builder, numElems: int) -> int:
    return MasterNodeUsageStartMrealmIdVector(builder, numElems)


def MasterNodeUsageAddTimestampFrom(builder, timestampFrom):
    builder.PrependUint64Slot(2, timestampFrom, 0)


def AddTimestampFrom(builder, timestampFrom):
    MasterNodeUsageAddTimestampFrom(builder, timestampFrom)


def MasterNodeUsageAddPubkey(builder, pubkey):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(pubkey), 0)


def AddPubkey(builder, pubkey):
    MasterNodeUsageAddPubkey(builder, pubkey)


def MasterNodeUsageStartPubkeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPubkeyVector(builder, numElems: int) -> int:
    return MasterNodeUsageStartPubkeyVector(builder, numElems)


def MasterNodeUsageAddClientIpAddress(builder, clientIpAddress):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(clientIpAddress), 0)


def AddClientIpAddress(builder, clientIpAddress):
    MasterNodeUsageAddClientIpAddress(builder, clientIpAddress)


def MasterNodeUsageStartClientIpAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartClientIpAddressVector(builder, numElems: int) -> int:
    return MasterNodeUsageStartClientIpAddressVector(builder, numElems)


def MasterNodeUsageAddClientIpVersion(builder, clientIpVersion):
    builder.PrependUint8Slot(5, clientIpVersion, 0)


def AddClientIpVersion(builder, clientIpVersion):
    MasterNodeUsageAddClientIpVersion(builder, clientIpVersion)


def MasterNodeUsageAddClientIpPort(builder, clientIpPort):
    builder.PrependUint16Slot(6, clientIpPort, 0)


def AddClientIpPort(builder, clientIpPort):
    MasterNodeUsageAddClientIpPort(builder, clientIpPort)


def MasterNodeUsageAddSeq(builder, seq):
    builder.PrependUint64Slot(7, seq, 0)


def AddSeq(builder, seq):
    MasterNodeUsageAddSeq(builder, seq)


def MasterNodeUsageAddSent(builder, sent):
    builder.PrependUint64Slot(8, sent, 0)


def AddSent(builder, sent):
    MasterNodeUsageAddSent(builder, sent)


def MasterNodeUsageAddProcessed(builder, processed):
    builder.PrependUint64Slot(9, processed, 0)


def AddProcessed(builder, processed):
    MasterNodeUsageAddProcessed(builder, processed)


def MasterNodeUsageAddStatus(builder, status):
    builder.PrependUint8Slot(10, status, 0)


def AddStatus(builder, status):
    MasterNodeUsageAddStatus(builder, status)


def MasterNodeUsageAddStatusMessage(builder, statusMessage):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(statusMessage), 0)


def AddStatusMessage(builder, statusMessage):
    MasterNodeUsageAddStatusMessage(builder, statusMessage)


def MasterNodeUsageAddMeteringId(builder, meteringId):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(meteringId), 0)


def AddMeteringId(builder, meteringId):
    MasterNodeUsageAddMeteringId(builder, meteringId)


def MasterNodeUsageStartMeteringIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMeteringIdVector(builder, numElems: int) -> int:
    return MasterNodeUsageStartMeteringIdVector(builder, numElems)


def MasterNodeUsageAddCount(builder, count):
    builder.PrependUint64Slot(13, count, 0)


def AddCount(builder, count):
    MasterNodeUsageAddCount(builder, count)


def MasterNodeUsageAddTotal(builder, total):
    builder.PrependUint64Slot(14, total, 0)


def AddTotal(builder, total):
    MasterNodeUsageAddTotal(builder, total)


def MasterNodeUsageAddNodes(builder, nodes):
    builder.PrependUint64Slot(15, nodes, 0)


def AddNodes(builder, nodes):
    MasterNodeUsageAddNodes(builder, nodes)


def MasterNodeUsageAddControllers(builder, controllers):
    builder.PrependUint64Slot(16, controllers, 0)


def AddControllers(builder, controllers):
    MasterNodeUsageAddControllers(builder, controllers)


def MasterNodeUsageAddHostmonitors(builder, hostmonitors):
    builder.PrependUint64Slot(17, hostmonitors, 0)


def AddHostmonitors(builder, hostmonitors):
    MasterNodeUsageAddHostmonitors(builder, hostmonitors)


def MasterNodeUsageAddRouters(builder, routers):
    builder.PrependUint64Slot(18, routers, 0)


def AddRouters(builder, routers):
    MasterNodeUsageAddRouters(builder, routers)


def MasterNodeUsageAddContainers(builder, containers):
    builder.PrependUint64Slot(19, containers, 0)


def AddContainers(builder, containers):
    MasterNodeUsageAddContainers(builder, containers)


def MasterNodeUsageAddGuests(builder, guests):
    builder.PrependUint64Slot(20, guests, 0)


def AddGuests(builder, guests):
    MasterNodeUsageAddGuests(builder, guests)


def MasterNodeUsageAddProxies(builder, proxies):
    builder.PrependUint64Slot(21, proxies, 0)


def AddProxies(builder, proxies):
    MasterNodeUsageAddProxies(builder, proxies)


def MasterNodeUsageAddMarketmakers(builder, marketmakers):
    builder.PrependUint64Slot(22, marketmakers, 0)


def AddMarketmakers(builder, marketmakers):
    MasterNodeUsageAddMarketmakers(builder, marketmakers)


def MasterNodeUsageAddSessions(builder, sessions):
    builder.PrependUint64Slot(23, sessions, 0)


def AddSessions(builder, sessions):
    MasterNodeUsageAddSessions(builder, sessions)


def MasterNodeUsageAddMsgsCall(builder, msgsCall):
    builder.PrependUint64Slot(24, msgsCall, 0)


def AddMsgsCall(builder, msgsCall):
    MasterNodeUsageAddMsgsCall(builder, msgsCall)


def MasterNodeUsageAddMsgsYield(builder, msgsYield):
    builder.PrependUint64Slot(25, msgsYield, 0)


def AddMsgsYield(builder, msgsYield):
    MasterNodeUsageAddMsgsYield(builder, msgsYield)


def MasterNodeUsageAddMsgsInvocation(builder, msgsInvocation):
    builder.PrependUint64Slot(26, msgsInvocation, 0)


def AddMsgsInvocation(builder, msgsInvocation):
    MasterNodeUsageAddMsgsInvocation(builder, msgsInvocation)


def MasterNodeUsageAddMsgsResult(builder, msgsResult):
    builder.PrependUint64Slot(27, msgsResult, 0)


def AddMsgsResult(builder, msgsResult):
    MasterNodeUsageAddMsgsResult(builder, msgsResult)


def MasterNodeUsageAddMsgsError(builder, msgsError):
    builder.PrependUint64Slot(28, msgsError, 0)


def AddMsgsError(builder, msgsError):
    MasterNodeUsageAddMsgsError(builder, msgsError)


def MasterNodeUsageAddMsgsPublish(builder, msgsPublish):
    builder.PrependUint64Slot(29, msgsPublish, 0)


def AddMsgsPublish(builder, msgsPublish):
    MasterNodeUsageAddMsgsPublish(builder, msgsPublish)


def MasterNodeUsageAddMsgsPublished(builder, msgsPublished):
    builder.PrependUint64Slot(30, msgsPublished, 0)


def AddMsgsPublished(builder, msgsPublished):
    MasterNodeUsageAddMsgsPublished(builder, msgsPublished)


def MasterNodeUsageAddMsgsEvent(builder, msgsEvent):
    builder.PrependUint64Slot(31, msgsEvent, 0)


def AddMsgsEvent(builder, msgsEvent):
    MasterNodeUsageAddMsgsEvent(builder, msgsEvent)


def MasterNodeUsageAddMsgsRegister(builder, msgsRegister):
    builder.PrependUint64Slot(32, msgsRegister, 0)


def AddMsgsRegister(builder, msgsRegister):
    MasterNodeUsageAddMsgsRegister(builder, msgsRegister)


def MasterNodeUsageAddMsgsRegistered(builder, msgsRegistered):
    builder.PrependUint64Slot(33, msgsRegistered, 0)


def AddMsgsRegistered(builder, msgsRegistered):
    MasterNodeUsageAddMsgsRegistered(builder, msgsRegistered)


def MasterNodeUsageAddMsgsSubscribe(builder, msgsSubscribe):
    builder.PrependUint64Slot(34, msgsSubscribe, 0)


def AddMsgsSubscribe(builder, msgsSubscribe):
    MasterNodeUsageAddMsgsSubscribe(builder, msgsSubscribe)


def MasterNodeUsageAddMsgsSubscribed(builder, msgsSubscribed):
    builder.PrependUint64Slot(35, msgsSubscribed, 0)


def AddMsgsSubscribed(builder, msgsSubscribed):
    MasterNodeUsageAddMsgsSubscribed(builder, msgsSubscribed)


def MasterNodeUsageEnd(builder):
    return builder.EndObject()


def End(builder):
    return MasterNodeUsageEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork


class VerificationStatus(object):
    # Unset
    NONE = 0
    #
    PENDING = 1
    #
    VERIFIED = 2
    #
    FAILED = 3
    #
    EXPIRED = 4



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.log.IpVersion import IpVersion
from cfxdb.gen.log.MasterNodeUsageStatus import MasterNodeUsageStatus
from cfxdb.gen.log.MasterState import MasterState
from cfxdb.gen.log.MNodeState import MNodeState
from cfxdb.gen.log.MRealmState import MRealmState
from cfxdb.gen.log.MWorkerState import MWorkerState
from cfxdb.gen.log.MWorkerType import MWorkerType
from cfxdb.log.mnode_log import MNodeLog
from cfxdb.log.mnode_logs import MNodeLogs
from cfxdb.log.mworker_log import MWorkerLog
from cfxdb.log.mworker_logs import MWorkerLogs
from cfxdb.log.schema import Schema

__all__ = (
    "MNodeLog",
    "MNodeLogs",
    "MWorkerLog",
    "MWorkerLogs",
    "Schema",
    "IpVersion",
    "MasterNodeUsageStatus",
    "MasterState",
    "MNodeState",
    "MRealmState",
    "MWorkerState",
    "MWorkerType",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR Market Actors.
class Actor(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Actor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsActor(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Actor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Ethereum address of the member.
    # Actor
    def Actor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def ActorAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def ActorLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def ActorIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Type of the market actor.
    # Actor
    def ActorType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # ID of the market this actor is associated with.
    # Actor
    def Market(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def MarketAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def MarketLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def MarketIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Database transaction time (epoch time in ns) of insert or last update.
    # Actor
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number (on the blockchain) when the actor (originally) joined the market.
    # Actor
    def Joined(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def JoinedAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def JoinedLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def JoinedIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Security (XBR tokens) deposited by the actor in the market.
    # Actor
    def Security(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def SecurityAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def SecurityLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def SecurityIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
    # Actor
    def Meta(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Actor
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # When signed off-chain and submitted via ``XBRNetwork.registerMemberFor``.
    # Actor
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Actor
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Actor
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Actor
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0


def ActorStart(builder):
    builder.StartObject(9)


def Start(builder):
    ActorStart(builder)


def ActorAddActor(builder, actor):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(actor), 0)


def AddActor(builder, actor):
    ActorAddActor(builder, actor)


def ActorStartActorVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartActorVector(builder, numElems: int) -> int:
    return ActorStartActorVector(builder, numElems)


def ActorAddActorType(builder, actorType):
    builder.PrependUint8Slot(1, actorType, 0)


def AddActorType(builder, actorType):
    ActorAddActorType(builder, actorType)


def ActorAddMarket(builder, market):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(market), 0)


def AddMarket(builder, market):
    ActorAddMarket(builder, market)


def ActorStartMarketVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketVector(builder, numElems: int) -> int:
    return ActorStartMarketVector(builder, numElems)


def ActorAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(3, timestamp, 0)


def AddTimestamp(builder, timestamp):
    ActorAddTimestamp(builder, timestamp)


def ActorAddJoined(builder, joined):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(joined), 0)


def AddJoined(builder, joined):
    ActorAddJoined(builder, joined)


def ActorStartJoinedVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartJoinedVector(builder, numElems: int) -> int:
    return ActorStartJoinedVector(builder, numElems)


def ActorAddSecurity(builder, security):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(security), 0)


def AddSecurity(builder, security):
    ActorAddSecurity(builder, security)


def ActorStartSecurityVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSecurityVector(builder, numElems: int) -> int:
    return ActorStartSecurityVector(builder, numElems)


def ActorAddMeta(builder, meta):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(meta), 0)


def AddMeta(builder, meta):
    ActorAddMeta(builder, meta)


def ActorAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    ActorAddTid(builder, tid)


def ActorStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return ActorStartTidVector(builder, numElems)


def ActorAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    ActorAddSignature(builder, signature)


def ActorStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return ActorStartSignatureVector(builder, numElems)


def ActorEnd(builder):
    return builder.EndObject()


def End(builder):
    return ActorEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapUuidBytes20Uint8FlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Actor as ActorGen


class _ActorGen(ActorGen.Actor):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsActor(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ActorGen()
        x.Init(buf, n + offset)
        return x

    def ActorAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MarketAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def JoinedAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SecurityAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Actor(object):
    """
    XBR Market Actors.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (address)
        self._actor = None

        # uint8 (ActorType)
        self._actor_type = None

        # [uint8] (uuid)
        self._market = None

        # uint64 (timestamp)
        self._timestamp = None

        # [uint8] (uint256)
        self._joined = None

        # [uint8] (uint256)
        self._security = None

        # string (multihash)
        self._meta = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

    def marshal(self):
        obj = {
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "market": self.market.bytes if self.market else None,
            "actor": bytes(self.actor) if self.actor else None,
            "actor_type": self.actor_type,
            "joined": pack_uint256(self.joined) if self.joined else None,
            "security": pack_uint256(self.security) if self.security else None,
            "meta": bytes(self.meta) if self.meta else None,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def market(self) -> uuid.UUID:
        """
        ID of the market this actor is associated with.
        """
        if self._market is None and self._from_fbs:
            if self._from_fbs.MarketLength():
                _market = self._from_fbs.MarketAsBytes()
                self._market = uuid.UUID(bytes=bytes(_market))
        return self._market

    @market.setter
    def market(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._market = value

    @property
    def actor(self) -> bytes:
        """
        Ethereum address of the member.
        """
        if self._actor is None and self._from_fbs:
            if self._from_fbs.ActorLength():
                self._actor = self._from_fbs.ActorAsBytes()
        return self._actor

    @actor.setter
    def actor(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._actor = value

    @property
    def actor_type(self) -> int:
        """
        Type of the market actor.
        """
        if self._actor_type is None and self._from_fbs:
            self._actor_type = self._from_fbs.ActorType()
        return self._actor_type

    @actor_type.setter
    def actor_type(self, value: int):
        assert value is None or type(value) == int
        self._actor_type = value

    @property
    def joined(self) -> int:
        """
        Block number (on the blockchain) when the actor (originally) joined the market.
        """
        if self._joined is None and self._from_fbs:
            if self._from_fbs.JoinedLength():
                _joined = self._from_fbs.JoinedAsBytes()
                self._joined = unpack_uint256(bytes(_joined))
            else:
                self._joined = 0
        return self._joined

    @joined.setter
    def joined(self, value: int):
        assert value is None or type(value) == int
        self._joined = value

    @property
    def security(self) -> int:
        """
        Security (XBR tokens) deposited by the actor in the market.
        """
        if self._security is None and self._from_fbs:
            if self._from_fbs.SecurityLength():
                security = self._from_fbs.SecurityAsBytes()
                self._security = unpack_uint256(bytes(security))
            else:
                self._security = 0
        return self._security

    @security.setter
    def security(self, value: int):
        assert value is None or type(value) == int, "security must be int, was: {}".format(value)
        self._security = value

    @property
    def meta(self) -> str:
        """
        The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
        """
        if self._meta is None and self._from_fbs:
            meta = self._from_fbs.Meta()
            if meta:
                self._meta = meta.decode("utf8")
        return self._meta

    @meta.setter
    def meta(self, value: str):
        assert value is None or type(value) == str
        self._meta = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRMarket.joinMarketFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return Actor(_ActorGen.GetRootAsActor(buf, 0))

    def build(self, builder):
        market = self.market.bytes if self.market else None
        if market:
            market = builder.CreateString(market)

        actor = self.actor
        if actor:
            actor = builder.CreateString(actor)

        joined = self.joined
        if joined:
            joined = builder.CreateString(pack_uint256(joined))

        security = self.security
        if security:
            security = builder.CreateString(pack_uint256(security))

        meta = self.meta
        if meta:
            meta = builder.CreateString(meta)

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        ActorGen.ActorStart(builder)

        if self.timestamp:
            ActorGen.ActorAddTimestamp(builder, int(self.timestamp))

        if market:
            ActorGen.ActorAddMarket(builder, market)

        if actor:
            ActorGen.ActorAddActor(builder, actor)

        if self.actor_type:
            ActorGen.ActorAddActorType(builder, self.actor_type)

        if joined:
            ActorGen.ActorAddJoined(builder, joined)

        if security:
            ActorGen.ActorAddSecurity(builder, security)

        if meta:
            ActorGen.ActorAddMeta(builder, meta)

        if tid:
            ActorGen.ActorAddTid(builder, tid)

        if signature:
            ActorGen.ActorAddSignature(builder, signature)

        final = ActorGen.ActorEnd(builder)

        return final


@table("1863eb64-322a-42dd-9fce-a59c99d5b40e", build=Actor.build, cast=Actor.cast)
class Actors(MapUuidBytes20Uint8FlatBuffers):
    """
    Market actors table, mapping from ``(market_id|UUID, actor_adr|bytes[20], actor_type|int)`` to :class:`cfxdb.xbr.Actor`.
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from zlmdb import (
    MapStringStringStringUuid,
    MapStringUuid,
    MapUuidCbor,
    MapUuidStringUuid,
    MapUuidTimestampUuid,
    MapUuidUuidCbor,
    MapUuidUuidUuid,
    MapUuidUuidUuidStringUuid,
    MapUuidUuidUuidUuid,
    table,
)

from cfxdb.log import MNodeLogs, MWorkerLogs
from cfxdb.mrealm import (
    ApplicationRealm,
    ApplicationRealmRoleAssociation,
    Credential,
    Permission,
    Principal,
    Role,
    RouterCluster,
    RouterClusterNodeMembership,
    RouterWorkerGroup,
    RouterWorkerGroupClusterPlacement,
    WebCluster,
    WebClusterNodeMembership,
    WebService,
)

__all__ = ("MrealmSchema",)


#
# Application Realms
#
@table("7099565b-7b44-4891-a0c8-83c7dbb60883", marshal=ApplicationRealm.marshal, parse=ApplicationRealm.parse)
class ApplicationRealms(MapUuidCbor):
    """
    Application realms defined for user application routing (``arealm_oid -> arealm``).
    """


@table("89f3073a-32d5-497e-887d-7e930e9c26e6")
class IndexApplicationRealmByName(MapStringUuid):
    """
    Index of application realms by realm name (``arealm_name -> arealm_oid``).
    """


@table("0275b858-890c-4879-945c-720235b093d7")
class IndexApplicationRealmByWebCluster(MapUuidStringUuid):
    """
    Index of application realms by webclusters and realm names (``(webcluster_oid, arealm_name) -> arealm_oid``).
    """


#
# Principals
#
@table("9808cb0b-1b55-4b3f-858e-39004cb11135", marshal=Principal.marshal, parse=Principal.parse)
class Principals(MapUuidCbor):
    """
    Principals created for use with WAMP authentication (``principal_oid -> principal``).
    """


@table("212f3455-6d4c-43ec-843d-53cd17e31974")
class IndexPrincipalByName(MapUuidStringUuid):
    """
    Index of principals by application realms and principal names (``(arealm_oid, principal_name) -> principal_oid``).
    """


#
# Credentials
#
@table("251c8620-425a-4eeb-ade9-4284e8670080", marshal=Credential.marshal, parse=Credential.parse)
class Credentials(MapUuidCbor):
    """
    Credentials created for use with WAMP authentication (``credential_oid -> credential``).
    """


@table("45490b42-b167-4df6-ab1c-41d434390397")
class IndexCredentialsByAuth(MapStringStringStringUuid):
    """
    Index of credentials by authentication method, realm name and authentication ID (``(authmethod, realm_name, authid) -> credential_oid``).
    """


@table("ed0da771-e331-4d93-b50c-d371391cd7b9")
class IndexCredentialsByPrincipal(MapUuidTimestampUuid):
    """
    Index of credentials by principals and modification date (``(principal_oid, modified) -> credential_oid``).
    """


#
# Roles
#
@table("341083bb-edeb-461c-a6d4-38dddcda6ec9", marshal=Role.marshal, parse=Role.parse)
class Roles(MapUuidCbor):
    """
    Roles created for use with application-level authorization and permissions in application realms (``role_oid -> role``).
    """


@table("71b990d1-4525-44cd-9ef8-3569de8b4c80")
class IndexRoleByName(MapStringUuid):
    """
    Index of roles by role names (``role_name -> role_oid``).
    """


#
# Permissions
#
@table("f98ed35b-f8fb-47ba-81e1-3c014101464d", marshal=Permission.marshal, parse=Permission.parse)
class Permissions(MapUuidCbor):
    """
    Role permission database object (``permission_oid -> permission``).
    """


@table("6cdc21bf-353d-4477-8631-8eb039142ae9")
class IndexPermissionByUri(MapUuidStringUuid):
    """
    Index of role permissions by roles and URIs (``(role_oid, uri) -> permission_oid``).
    """


#
# Application Realm Role Associations
#
@table(
    "5eabdb63-9c31-4c97-b514-7e8fbac7e143",
    marshal=ApplicationRealmRoleAssociation.marshal,
    parse=ApplicationRealmRoleAssociation.parse,
)
class ApplicationRealmRoleAssociations(MapUuidUuidCbor):
    """
    Association of a role with an application realm (``(arealm_oid, role_oid) -> arealm_role_association``).
    """


#
# Router clusters
#
@table("b054a230-c370-4c29-b5de-7e0148321b0a", marshal=RouterCluster.marshal, parse=RouterCluster.parse)
class RouterClusters(MapUuidCbor):
    """
    A router cluster is able to run (application) realms, and is hosted on a groups of router workers
    kept in sync and meshed via router-to-router links (``routercluster_oid -> routercluster``).
    """


@table("0c80c7a8-7536-4a74-8916-4922c0b72cb7")
class IndexRouterClusterByName(MapStringUuid):
    """
    Index of router clusters by router cluster names (``routercluster_name -> routercluster_oid``).
    """


#
# Router cluster node memberships
#
@table(
    "a091bad6-f14c-437c-8e30-e9be84380658",
    marshal=RouterClusterNodeMembership.marshal,
    parse=RouterClusterNodeMembership.parse,
)
class RouterClusterNodeMemberships(MapUuidUuidCbor):
    """
    Membership of a managed node (paired within a management realm) to a router cluster (``(cluster_oid, node_oid) -> cluster_node_membership``).
    """


#
# Router worker groups
#
@table("c019457b-d499-454f-9bf2-4f7e85079d8f", marshal=RouterWorkerGroup.marshal, parse=RouterWorkerGroup.parse)
class RouterWorkerGroups(MapUuidCbor):
    """
    Router worker group database configuration object (``workergroup_oid -> workergroup``).
    """


@table("4bb8ec14-4820-4061-8b2c-d1841e2686e1")
class IndexWorkerGroupByCluster(MapUuidStringUuid):
    """
    Index of router worker groups by router clusters and group names (``(cluster_oid, workergroup_name) -> workergroup_oid``).
    """


@table("4c7d184b-2303-492d-822d-ed12516050a9")
class IndexWorkerGroupByPlacement(MapUuidUuidUuidUuid):
    """
    Index of router worker groups by cluster, node and placement (``(cluster_oid, node_oid, placement_oid) -> workergroup_oid``).
    """


#
# Router worker groups to cluster node placements
#
@table(
    "e3d326d2-6140-47a9-adf9-8e93b832717b",
    marshal=RouterWorkerGroupClusterPlacement.marshal,
    parse=RouterWorkerGroupClusterPlacement.parse,
)
class RouterWorkerGroupClusterPlacements(MapUuidCbor):
    """
    Router worker group placements of workers (``placement_oid -> placement``).
    """


@table("1a18739f-7224-4459-a446-6f1fedd760a7")
class IndexClusterPlacementByWorkerName(MapUuidUuidUuidStringUuid):
    """
    Index of router worker group placements by worker group, cluster, node and worker name (``(workergroup_oid, cluster_oid, node_oid, worker_name) -> placement_oid``).
    """


#
# Web clusters
#
@table("719d029f-e9d5-4b25-98e0-cf04d5a2648b", marshal=WebCluster.marshal, parse=WebCluster.parse)
class WebClusters(MapUuidCbor):
    """
    Web cluster database configuration object (``webcluster_oid -> webcluster``).
    """


@table("296c7d17-4769-4e40-8cb7-e6c394b93335")
class IndexWebClusterByName(MapStringUuid):
    """
    Index of web clusters by cluster name (``webcluster_name -> webcluster_oid``).
    """


#
# Web cluster node memberships
#
@table(
    "e9801077-a629-470b-a4c9-4292a1f00d43",
    marshal=WebClusterNodeMembership.marshal,
    parse=WebClusterNodeMembership.parse,
)
class WebClusterNodeMemberships(MapUuidUuidCbor):
    """
    Information about memberships of nodes in web clusters (``(webcluster_oid, node_oid) -> webcluster_node_membership``).
    """


#
# Web cluster services
#
@table("a8803ca3-09a0-4d72-8728-2469de8d50ac", marshal=WebService.marshal, parse=WebService.parse)
class WebServices(MapUuidCbor):
    """
    Web cluster services (``webservice_oid -> webservice``).
    """


@table("d23d4dbb-5d5c-4ccc-b72a-0ff18363169f")
class IndexWebClusterWebServices(MapUuidUuidUuid):
    """
    Index of web services by web cluster and web service (``(webcluster_oid, webservice_oid) -> webservice_oid``).
    """


@table("f0b05bcf-f682-49bb-929e-ac252e9867fa")
class IndexWebServiceByPath(MapUuidStringUuid):
    """
    Index of web services by web cluster and web service name (``(webcluster_oid, webservice_name) -> webservice_oid``).
    """


@table("62d0841c-602e-473e-a6d5-3d8ce01e9e06")
class IndexWebClusterPathToWebService(MapUuidStringUuid):
    """
    Index of web services by web cluster and web path (``(webcluster_oid, path) -> webservice_oid``).
    """


class MrealmSchema(object):
    """
    Management realm database schema.
    """

    def __init__(self, db):
        self.db = db

    principals: Principals
    """
    Application realm client principals.
    """

    idx_principals_by_name: IndexPrincipalByName
    """
    Index on principals (by name).
    """

    credentials: Credentials
    """
    WAMP client authentication credentials, used for mapping ``(authmethod, realm, authid) -> principal``.
    """

    idx_credentials_by_auth: IndexCredentialsByAuth
    """
    Index on credentials (by WAMP auth information).
    """

    idx_credentials_by_principal: IndexCredentialsByPrincipal
    """
    Index on credentials (by principal_oid, modified).
    """

    roles: Roles
    """
    Roles for used in authorization with application routing.
    """

    idx_roles_by_name: IndexRoleByName
    """
    Index on roles (by name).
    """

    permissions: Permissions
    """
    Permissions defined on roles.
    """

    idx_permissions_by_uri: IndexPermissionByUri
    """
    Index on permissions: by URI.
    """

    arealms: ApplicationRealms
    """
    Application realms defined in this management realm.
    """

    idx_arealms_by_name: IndexApplicationRealmByName
    """
    Index on application realms: by name.
    """

    idx_arealm_by_webcluster: IndexApplicationRealmByWebCluster
    """
    Index on application realms: by web cluster.
    """

    arealm_role_associations: ApplicationRealmRoleAssociations
    """
    Association of roles to application realms.
    """

    routerclusters: RouterClusters
    """
    Router clusters defined in this management realm.
    """

    idx_routerclusters_by_name: IndexRouterClusterByName
    """
    Index on router clusters: by name.
    """

    routercluster_node_memberships: RouterClusterNodeMemberships
    """
    Node membership in router clusters.
    """

    router_workergroups: RouterWorkerGroups
    """
    Router worker groups.
    """

    idx_workergroup_by_cluster: IndexWorkerGroupByCluster
    """
    Index on worker groups: by cluster.
    """

    idx_workergroup_by_placement: IndexWorkerGroupByPlacement
    """
    Index on worker groups: by placement.
    """

    router_workergroup_placements: RouterWorkerGroupClusterPlacements
    """
    Router worker cluster placements.
    """

    idx_clusterplacement_by_workername: IndexClusterPlacementByWorkerName
    """
    Index on router worker placements: by worker name.
    """

    webclusters: WebClusters
    """
    Web clusters.
    """

    idx_webclusters_by_name: IndexWebClusterByName
    """
    Index of web clusters: by name.
    """

    webcluster_node_memberships: WebClusterNodeMemberships
    """
    Node membership in web clusters.
    """

    webservices: WebServices
    """
    Web service added to web clusters.
    """

    idx_webservices_by_path: IndexWebServiceByPath
    """
    Index on web services: by HTTP path.
    """

    idx_webcluster_webservices: IndexWebClusterWebServices
    """
    Index on web service: by ...
    """

    mnode_logs: MNodeLogs
    """
    Managed node log records.
    """

    mworker_logs: MWorkerLogs
    """
    Managed node worker log records.
    """

    @staticmethod
    def attach(db):
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: zlmdb.Database
        :return: object of Schema
        """
        schema = MrealmSchema(db)

        # application realms
        schema.arealms = db.attach_table(ApplicationRealms)

        schema.idx_arealms_by_name = db.attach_table(IndexApplicationRealmByName)
        schema.arealms.attach_index("idx1", schema.idx_arealms_by_name, lambda arealm: arealm.name)

        schema.idx_arealm_by_webcluster = db.attach_table(IndexApplicationRealmByWebCluster)
        schema.arealms.attach_index(
            "idx2", schema.idx_arealm_by_webcluster, lambda arealm: (arealm.webcluster_oid, arealm.name), nullable=True
        )

        # principals
        schema.principals = db.attach_table(Principals)

        schema.idx_principals_by_name = db.attach_table(IndexPrincipalByName)
        schema.principals.attach_index(
            "idx1", schema.idx_principals_by_name, lambda principal: (principal.arealm_oid, principal.authid)
        )

        # credentials
        schema.credentials = db.attach_table(Credentials)

        schema.idx_credentials_by_auth = db.attach_table(IndexCredentialsByAuth)
        schema.credentials.attach_index(
            "idx1",
            schema.idx_credentials_by_auth,
            lambda credential: (credential.authmethod, credential.realm, credential.authid),
        )

        schema.idx_credentials_by_principal = db.attach_table(IndexCredentialsByPrincipal)
        schema.credentials.attach_index(
            "idx2",
            schema.idx_credentials_by_principal,
            lambda credential: (credential.principal_oid, credential.created),
        )

        # roles
        schema.roles = db.attach_table(Roles)

        schema.idx_roles_by_name = db.attach_table(IndexRoleByName)
        schema.roles.attach_index("idx1", schema.idx_roles_by_name, lambda role: role.name)

        schema.arealm_role_associations = db.attach_table(ApplicationRealmRoleAssociations)

        # permissions
        schema.permissions = db.attach_table(Permissions)

        schema.idx_permissions_by_uri = db.attach_table(IndexPermissionByUri)
        schema.permissions.attach_index(
            "idx1", schema.idx_permissions_by_uri, lambda permission: (permission.role_oid, permission.uri)
        )

        # router clusters
        schema.routerclusters = db.attach_table(RouterClusters)

        schema.idx_routerclusters_by_name = db.attach_table(IndexRouterClusterByName)
        schema.routerclusters.attach_index(
            "idx1", schema.idx_routerclusters_by_name, lambda routercluster: routercluster.name
        )

        schema.routercluster_node_memberships = db.attach_table(RouterClusterNodeMemberships)

        # router worker groups
        schema.router_workergroups = db.attach_table(RouterWorkerGroups)

        schema.idx_workergroup_by_cluster = db.attach_table(IndexWorkerGroupByCluster)
        schema.router_workergroups.attach_index(
            "idx1", schema.idx_workergroup_by_cluster, lambda wg: (wg.cluster_oid, wg.name)
        )

        # router worker group placements
        schema.router_workergroup_placements = db.attach_table(RouterWorkerGroupClusterPlacements)

        # index: (workergroup_oid, cluster_oid, node_oid, worker_name) -> placement_oid
        schema.idx_clusterplacement_by_workername = db.attach_table(IndexClusterPlacementByWorkerName)
        schema.router_workergroup_placements.attach_index(
            "idx1",
            schema.idx_clusterplacement_by_workername,
            lambda p: (p.worker_group_oid, p.cluster_oid, p.node_oid, p.worker_name),
        )

        # index: (cluster_oid, node_oid, placement_oid) -> placement_oid
        schema.idx_workergroup_by_placement = db.attach_table(IndexWorkerGroupByPlacement)
        schema.router_workergroup_placements.attach_index(
            "idx2", schema.idx_workergroup_by_placement, lambda p: (p.cluster_oid, p.node_oid, p.oid)
        )

        # web clusters
        schema.webclusters = db.attach_table(WebClusters)

        schema.idx_webclusters_by_name = db.attach_table(IndexWebClusterByName)
        schema.webclusters.attach_index("idx1", schema.idx_webclusters_by_name, lambda webcluster: webcluster.name)

        schema.webcluster_node_memberships = db.attach_table(WebClusterNodeMemberships)

        # web services
        schema.webservices = db.attach_table(WebServices)

        schema.idx_webservices_by_path = db.attach_table(IndexWebServiceByPath)
        schema.webservices.attach_index(
            "idx1", schema.idx_webservices_by_path, lambda webservice: (webservice.webcluster_oid, webservice.path)
        )

        schema.idx_webcluster_webservices = db.attach_table(IndexWebClusterWebServices)
        schema.webservices.attach_index(
            "idx2", schema.idx_webcluster_webservices, lambda webservice: (webservice.webcluster_oid, webservice.oid)
        )

        schema.mnode_logs = db.attach_table(MNodeLogs)
        schema.mworker_logs = db.attach_table(MWorkerLogs)

        return schema



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Heartbeat log record from managed node workers. Primary key: (timestamp, node_id, worker_id).
class MWorkerLog(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MWorkerLog()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMWorkerLog(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # MWorkerLog
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: Unix time in ns when this log record was received (from CFC node clock).
    # MWorkerLog
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Primary key: CF node ID.
    # MWorkerLog
    def NodeId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MWorkerLog
    def NodeIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MWorkerLog
    def NodeIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MWorkerLog
    def NodeIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Primary key: Local Worker ID on running CF node.
    # MWorkerLog
    def WorkerId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Heartbeat period in seconds encompassed by this record.
    # MWorkerLog
    def Period(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Management realm ID.
    # MWorkerLog
    def MrealmId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MWorkerLog
    def MrealmIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MWorkerLog
    def MrealmIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MWorkerLog
    def MrealmIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Type of worker.
    # MWorkerLog
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Current state of worker.
    # MWorkerLog
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Unix time in ns. This timestamp is from the original received event payload (from CF node clock).
    # MWorkerLog
    def Sent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Sequence number as sent in the log record by the CF node (started at 0 for CF start and incremented by one on each heartbeat).
    # MWorkerLog
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Current load.
    # MWorkerLog
    def NumFds(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def NumThreads(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def NumCtxSwitchesInvoluntary(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def NumCtxSwitchesInvoluntaryPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def NumCtxSwitchesVoluntary(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def NumCtxSwitchesVoluntaryPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # CPU load
    # MWorkerLog
    def CpuSystem(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MWorkerLog
    def CpuSystemPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MWorkerLog
    def CpuWait(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MWorkerLog
    def CpuWaitPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MWorkerLog
    def CpuUser(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MWorkerLog
    def CpuUserPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # Disk load
    # MWorkerLog
    def ReadBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def ReadBytesPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def ReadIos(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(50))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def ReadIosPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(52))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def WriteBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(54))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def WriteBytesPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(56))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def WriteIos(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(58))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def WriteIosPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(60))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Network load
    # MWorkerLog
    def SentBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(62))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentBytesPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(64))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentIos(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(66))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentIosPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(68))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(70))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvBytesPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(72))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvIos(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(74))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvIosPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(76))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP roles defined on a router worker
    # MWorkerLog
    def RouterRoles(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(78))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP sessions joined on a router worker
    # MWorkerLog
    def RouterSessions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(80))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP messages processed by a router worker
    # MWorkerLog
    def RecvCall(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(82))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvYield(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(84))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentInvocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(86))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentResult(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(88))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvPublish(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(90))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentPublished(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(92))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentEvent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(94))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvRegister(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(96))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentRegistered(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(98))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def RecvSubscribe(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(100))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MWorkerLog
    def SentSubscribed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(102))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def MWorkerLogStart(builder):
    builder.StartObject(50)


def Start(builder):
    MWorkerLogStart(builder)


def MWorkerLogAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    MWorkerLogAddTimestamp(builder, timestamp)


def MWorkerLogAddNodeId(builder, nodeId):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(nodeId), 0)


def AddNodeId(builder, nodeId):
    MWorkerLogAddNodeId(builder, nodeId)


def MWorkerLogStartNodeIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartNodeIdVector(builder, numElems: int) -> int:
    return MWorkerLogStartNodeIdVector(builder, numElems)


def MWorkerLogAddWorkerId(builder, workerId):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(workerId), 0)


def AddWorkerId(builder, workerId):
    MWorkerLogAddWorkerId(builder, workerId)


def MWorkerLogAddPeriod(builder, period):
    builder.PrependUint32Slot(3, period, 0)


def AddPeriod(builder, period):
    MWorkerLogAddPeriod(builder, period)


def MWorkerLogAddMrealmId(builder, mrealmId):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(mrealmId), 0)


def AddMrealmId(builder, mrealmId):
    MWorkerLogAddMrealmId(builder, mrealmId)


def MWorkerLogStartMrealmIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMrealmIdVector(builder, numElems: int) -> int:
    return MWorkerLogStartMrealmIdVector(builder, numElems)


def MWorkerLogAddType(builder, type):
    builder.PrependUint8Slot(5, type, 0)


def AddType(builder, type):
    MWorkerLogAddType(builder, type)


def MWorkerLogAddState(builder, state):
    builder.PrependUint8Slot(6, state, 0)


def AddState(builder, state):
    MWorkerLogAddState(builder, state)


def MWorkerLogAddSent(builder, sent):
    builder.PrependUint64Slot(7, sent, 0)


def AddSent(builder, sent):
    MWorkerLogAddSent(builder, sent)


def MWorkerLogAddSeq(builder, seq):
    builder.PrependUint64Slot(8, seq, 0)


def AddSeq(builder, seq):
    MWorkerLogAddSeq(builder, seq)


def MWorkerLogAddNumFds(builder, numFds):
    builder.PrependUint32Slot(9, numFds, 0)


def AddNumFds(builder, numFds):
    MWorkerLogAddNumFds(builder, numFds)


def MWorkerLogAddNumThreads(builder, numThreads):
    builder.PrependUint16Slot(10, numThreads, 0)


def AddNumThreads(builder, numThreads):
    MWorkerLogAddNumThreads(builder, numThreads)


def MWorkerLogAddNumCtxSwitchesInvoluntary(builder, numCtxSwitchesInvoluntary):
    builder.PrependUint64Slot(11, numCtxSwitchesInvoluntary, 0)


def AddNumCtxSwitchesInvoluntary(builder, numCtxSwitchesInvoluntary):
    MWorkerLogAddNumCtxSwitchesInvoluntary(builder, numCtxSwitchesInvoluntary)


def MWorkerLogAddNumCtxSwitchesInvoluntaryPerSec(builder, numCtxSwitchesInvoluntaryPerSec):
    builder.PrependUint32Slot(12, numCtxSwitchesInvoluntaryPerSec, 0)


def AddNumCtxSwitchesInvoluntaryPerSec(builder, numCtxSwitchesInvoluntaryPerSec):
    MWorkerLogAddNumCtxSwitchesInvoluntaryPerSec(builder, numCtxSwitchesInvoluntaryPerSec)


def MWorkerLogAddNumCtxSwitchesVoluntary(builder, numCtxSwitchesVoluntary):
    builder.PrependUint64Slot(13, numCtxSwitchesVoluntary, 0)


def AddNumCtxSwitchesVoluntary(builder, numCtxSwitchesVoluntary):
    MWorkerLogAddNumCtxSwitchesVoluntary(builder, numCtxSwitchesVoluntary)


def MWorkerLogAddNumCtxSwitchesVoluntaryPerSec(builder, numCtxSwitchesVoluntaryPerSec):
    builder.PrependUint32Slot(14, numCtxSwitchesVoluntaryPerSec, 0)


def AddNumCtxSwitchesVoluntaryPerSec(builder, numCtxSwitchesVoluntaryPerSec):
    MWorkerLogAddNumCtxSwitchesVoluntaryPerSec(builder, numCtxSwitchesVoluntaryPerSec)


def MWorkerLogAddCpuSystem(builder, cpuSystem):
    builder.PrependFloat32Slot(15, cpuSystem, 0.0)


def AddCpuSystem(builder, cpuSystem):
    MWorkerLogAddCpuSystem(builder, cpuSystem)


def MWorkerLogAddCpuSystemPerSec(builder, cpuSystemPerSec):
    builder.PrependFloat32Slot(16, cpuSystemPerSec, 0.0)


def AddCpuSystemPerSec(builder, cpuSystemPerSec):
    MWorkerLogAddCpuSystemPerSec(builder, cpuSystemPerSec)


def MWorkerLogAddCpuWait(builder, cpuWait):
    builder.PrependFloat32Slot(17, cpuWait, 0.0)


def AddCpuWait(builder, cpuWait):
    MWorkerLogAddCpuWait(builder, cpuWait)


def MWorkerLogAddCpuWaitPerSec(builder, cpuWaitPerSec):
    builder.PrependFloat32Slot(18, cpuWaitPerSec, 0.0)


def AddCpuWaitPerSec(builder, cpuWaitPerSec):
    MWorkerLogAddCpuWaitPerSec(builder, cpuWaitPerSec)


def MWorkerLogAddCpuUser(builder, cpuUser):
    builder.PrependFloat32Slot(19, cpuUser, 0.0)


def AddCpuUser(builder, cpuUser):
    MWorkerLogAddCpuUser(builder, cpuUser)


def MWorkerLogAddCpuUserPerSec(builder, cpuUserPerSec):
    builder.PrependFloat32Slot(20, cpuUserPerSec, 0.0)


def AddCpuUserPerSec(builder, cpuUserPerSec):
    MWorkerLogAddCpuUserPerSec(builder, cpuUserPerSec)


def MWorkerLogAddReadBytes(builder, readBytes):
    builder.PrependUint64Slot(21, readBytes, 0)


def AddReadBytes(builder, readBytes):
    MWorkerLogAddReadBytes(builder, readBytes)


def MWorkerLogAddReadBytesPerSec(builder, readBytesPerSec):
    builder.PrependUint64Slot(22, readBytesPerSec, 0)


def AddReadBytesPerSec(builder, readBytesPerSec):
    MWorkerLogAddReadBytesPerSec(builder, readBytesPerSec)


def MWorkerLogAddReadIos(builder, readIos):
    builder.PrependUint64Slot(23, readIos, 0)


def AddReadIos(builder, readIos):
    MWorkerLogAddReadIos(builder, readIos)


def MWorkerLogAddReadIosPerSec(builder, readIosPerSec):
    builder.PrependUint64Slot(24, readIosPerSec, 0)


def AddReadIosPerSec(builder, readIosPerSec):
    MWorkerLogAddReadIosPerSec(builder, readIosPerSec)


def MWorkerLogAddWriteBytes(builder, writeBytes):
    builder.PrependUint64Slot(25, writeBytes, 0)


def AddWriteBytes(builder, writeBytes):
    MWorkerLogAddWriteBytes(builder, writeBytes)


def MWorkerLogAddWriteBytesPerSec(builder, writeBytesPerSec):
    builder.PrependUint64Slot(26, writeBytesPerSec, 0)


def AddWriteBytesPerSec(builder, writeBytesPerSec):
    MWorkerLogAddWriteBytesPerSec(builder, writeBytesPerSec)


def MWorkerLogAddWriteIos(builder, writeIos):
    builder.PrependUint64Slot(27, writeIos, 0)


def AddWriteIos(builder, writeIos):
    MWorkerLogAddWriteIos(builder, writeIos)


def MWorkerLogAddWriteIosPerSec(builder, writeIosPerSec):
    builder.PrependUint64Slot(28, writeIosPerSec, 0)


def AddWriteIosPerSec(builder, writeIosPerSec):
    MWorkerLogAddWriteIosPerSec(builder, writeIosPerSec)


def MWorkerLogAddSentBytes(builder, sentBytes):
    builder.PrependUint64Slot(29, sentBytes, 0)


def AddSentBytes(builder, sentBytes):
    MWorkerLogAddSentBytes(builder, sentBytes)


def MWorkerLogAddSentBytesPerSec(builder, sentBytesPerSec):
    builder.PrependUint64Slot(30, sentBytesPerSec, 0)


def AddSentBytesPerSec(builder, sentBytesPerSec):
    MWorkerLogAddSentBytesPerSec(builder, sentBytesPerSec)


def MWorkerLogAddSentIos(builder, sentIos):
    builder.PrependUint64Slot(31, sentIos, 0)


def AddSentIos(builder, sentIos):
    MWorkerLogAddSentIos(builder, sentIos)


def MWorkerLogAddSentIosPerSec(builder, sentIosPerSec):
    builder.PrependUint64Slot(32, sentIosPerSec, 0)


def AddSentIosPerSec(builder, sentIosPerSec):
    MWorkerLogAddSentIosPerSec(builder, sentIosPerSec)


def MWorkerLogAddRecvBytes(builder, recvBytes):
    builder.PrependUint64Slot(33, recvBytes, 0)


def AddRecvBytes(builder, recvBytes):
    MWorkerLogAddRecvBytes(builder, recvBytes)


def MWorkerLogAddRecvBytesPerSec(builder, recvBytesPerSec):
    builder.PrependUint64Slot(34, recvBytesPerSec, 0)


def AddRecvBytesPerSec(builder, recvBytesPerSec):
    MWorkerLogAddRecvBytesPerSec(builder, recvBytesPerSec)


def MWorkerLogAddRecvIos(builder, recvIos):
    builder.PrependUint64Slot(35, recvIos, 0)


def AddRecvIos(builder, recvIos):
    MWorkerLogAddRecvIos(builder, recvIos)


def MWorkerLogAddRecvIosPerSec(builder, recvIosPerSec):
    builder.PrependUint64Slot(36, recvIosPerSec, 0)


def AddRecvIosPerSec(builder, recvIosPerSec):
    MWorkerLogAddRecvIosPerSec(builder, recvIosPerSec)


def MWorkerLogAddRouterRoles(builder, routerRoles):
    builder.PrependUint64Slot(37, routerRoles, 0)


def AddRouterRoles(builder, routerRoles):
    MWorkerLogAddRouterRoles(builder, routerRoles)


def MWorkerLogAddRouterSessions(builder, routerSessions):
    builder.PrependUint64Slot(38, routerSessions, 0)


def AddRouterSessions(builder, routerSessions):
    MWorkerLogAddRouterSessions(builder, routerSessions)


def MWorkerLogAddRecvCall(builder, recvCall):
    builder.PrependUint64Slot(39, recvCall, 0)


def AddRecvCall(builder, recvCall):
    MWorkerLogAddRecvCall(builder, recvCall)


def MWorkerLogAddRecvYield(builder, recvYield):
    builder.PrependUint64Slot(40, recvYield, 0)


def AddRecvYield(builder, recvYield):
    MWorkerLogAddRecvYield(builder, recvYield)


def MWorkerLogAddSentInvocation(builder, sentInvocation):
    builder.PrependUint64Slot(41, sentInvocation, 0)


def AddSentInvocation(builder, sentInvocation):
    MWorkerLogAddSentInvocation(builder, sentInvocation)


def MWorkerLogAddSentResult(builder, sentResult):
    builder.PrependUint64Slot(42, sentResult, 0)


def AddSentResult(builder, sentResult):
    MWorkerLogAddSentResult(builder, sentResult)


def MWorkerLogAddRecvPublish(builder, recvPublish):
    builder.PrependUint64Slot(43, recvPublish, 0)


def AddRecvPublish(builder, recvPublish):
    MWorkerLogAddRecvPublish(builder, recvPublish)


def MWorkerLogAddSentPublished(builder, sentPublished):
    builder.PrependUint64Slot(44, sentPublished, 0)


def AddSentPublished(builder, sentPublished):
    MWorkerLogAddSentPublished(builder, sentPublished)


def MWorkerLogAddSentEvent(builder, sentEvent):
    builder.PrependUint64Slot(45, sentEvent, 0)


def AddSentEvent(builder, sentEvent):
    MWorkerLogAddSentEvent(builder, sentEvent)


def MWorkerLogAddRecvRegister(builder, recvRegister):
    builder.PrependUint64Slot(46, recvRegister, 0)


def AddRecvRegister(builder, recvRegister):
    MWorkerLogAddRecvRegister(builder, recvRegister)


def MWorkerLogAddSentRegistered(builder, sentRegistered):
    builder.PrependUint64Slot(47, sentRegistered, 0)


def AddSentRegistered(builder, sentRegistered):
    MWorkerLogAddSentRegistered(builder, sentRegistered)


def MWorkerLogAddRecvSubscribe(builder, recvSubscribe):
    builder.PrependUint64Slot(48, recvSubscribe, 0)


def AddRecvSubscribe(builder, recvSubscribe):
    MWorkerLogAddRecvSubscribe(builder, recvSubscribe)


def MWorkerLogAddSentSubscribed(builder, sentSubscribed):
    builder.PrependUint64Slot(49, sentSubscribed, 0)


def AddSentSubscribed(builder, sentSubscribed):
    MWorkerLogAddSentSubscribed(builder, sentSubscribed)


def MWorkerLogEnd(builder):
    return builder.EndObject()


def End(builder):
    return MWorkerLogEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Data encryption key off-chain offerings, submitted by seller/provider delegates in the market.
class Offer(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Offer()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsOffer(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Offer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Offer transaction time (epoch time in ns)
    # Offer
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # ID of the data encryption key offer.
    # Offer
    def Offer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def OfferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def OfferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def OfferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Address of the XBR provider offering the data encryption key.
    # Offer
    def Seller(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def SellerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def SellerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def SellerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # WAMP session ID of the caller that originally placed this offer.
    # Offer
    def SellerSessionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP session authid of the caller that originally placed this offer.
    # Offer
    def SellerAuthid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ID of the data encryption key offered.
    # Offer
    def Key(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def KeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def KeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def KeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # ID of the API the encrypted data (this key is for) is provided under.
    # Offer
    def Api(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def ApiAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def ApiLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def ApiIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # URI under which the data encrypted with the key offered is provided under.
    # Offer
    def Uri(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp from which the offer is valid (epoch time in ns).
    # Offer
    def ValidFrom(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Seller delegate signature for the offer. The signature covers all information of the original offer placement request and requestor.
    # Offer
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # Price of data encryption key in ERC20 tokens of the market coin type.
    # Offer
    def Price(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Offer
    def PriceAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Offer
    def PriceLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def PriceIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # Keys for optional user defined categories the specific data that is provided falls under.
    # Offer
    def CategoriesKey(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Offer
    def CategoriesKeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def CategoriesKeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # Values for optional user defined categories the specific data that is provided falls under.
    # Offer
    def CategoriesValue(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Offer
    def CategoriesValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Offer
    def CategoriesValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Optional data at which this offer expires (epoch time in ns).
    # Offer
    def Expires(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Optional maximum number of times this data encryption key is to be sold or 0 for unlimited.
    # Offer
    def Copies(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Remaining number of copies to be sold (if "copies" is set >0, otherwise 0).
    # Offer
    def Remaining(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0


def OfferStart(builder):
    builder.StartObject(16)


def Start(builder):
    OfferStart(builder)


def OfferAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    OfferAddTimestamp(builder, timestamp)


def OfferAddOffer(builder, offer):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(offer), 0)


def AddOffer(builder, offer):
    OfferAddOffer(builder, offer)


def OfferStartOfferVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOfferVector(builder, numElems: int) -> int:
    return OfferStartOfferVector(builder, numElems)


def OfferAddSeller(builder, seller):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(seller), 0)


def AddSeller(builder, seller):
    OfferAddSeller(builder, seller)


def OfferStartSellerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSellerVector(builder, numElems: int) -> int:
    return OfferStartSellerVector(builder, numElems)


def OfferAddSellerSessionId(builder, sellerSessionId):
    builder.PrependUint64Slot(3, sellerSessionId, 0)


def AddSellerSessionId(builder, sellerSessionId):
    OfferAddSellerSessionId(builder, sellerSessionId)


def OfferAddSellerAuthid(builder, sellerAuthid):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(sellerAuthid), 0)


def AddSellerAuthid(builder, sellerAuthid):
    OfferAddSellerAuthid(builder, sellerAuthid)


def OfferAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0)


def AddKey(builder, key):
    OfferAddKey(builder, key)


def OfferStartKeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKeyVector(builder, numElems: int) -> int:
    return OfferStartKeyVector(builder, numElems)


def OfferAddApi(builder, api):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(api), 0)


def AddApi(builder, api):
    OfferAddApi(builder, api)


def OfferStartApiVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartApiVector(builder, numElems: int) -> int:
    return OfferStartApiVector(builder, numElems)


def OfferAddUri(builder, uri):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(uri), 0)


def AddUri(builder, uri):
    OfferAddUri(builder, uri)


def OfferAddValidFrom(builder, validFrom):
    builder.PrependUint64Slot(8, validFrom, 0)


def AddValidFrom(builder, validFrom):
    OfferAddValidFrom(builder, validFrom)


def OfferAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    OfferAddSignature(builder, signature)


def OfferStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return OfferStartSignatureVector(builder, numElems)


def OfferAddPrice(builder, price):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(price), 0)


def AddPrice(builder, price):
    OfferAddPrice(builder, price)


def OfferStartPriceVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPriceVector(builder, numElems: int) -> int:
    return OfferStartPriceVector(builder, numElems)


def OfferAddCategoriesKey(builder, categoriesKey):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(categoriesKey), 0)


def AddCategoriesKey(builder, categoriesKey):
    OfferAddCategoriesKey(builder, categoriesKey)


def OfferStartCategoriesKeyVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartCategoriesKeyVector(builder, numElems: int) -> int:
    return OfferStartCategoriesKeyVector(builder, numElems)


def OfferAddCategoriesValue(builder, categoriesValue):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(categoriesValue), 0)


def AddCategoriesValue(builder, categoriesValue):
    OfferAddCategoriesValue(builder, categoriesValue)


def OfferStartCategoriesValueVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartCategoriesValueVector(builder, numElems: int) -> int:
    return OfferStartCategoriesValueVector(builder, numElems)


def OfferAddExpires(builder, expires):
    builder.PrependUint64Slot(13, expires, 0)


def AddExpires(builder, expires):
    OfferAddExpires(builder, expires)


def OfferAddCopies(builder, copies):
    builder.PrependUint32Slot(14, copies, 0)


def AddCopies(builder, copies):
    OfferAddCopies(builder, copies)


def OfferAddRemaining(builder, remaining):
    builder.PrependUint32Slot(15, remaining, 0)


def AddRemaining(builder, remaining):
    OfferAddRemaining(builder, remaining)


def OfferEnd(builder):
    return builder.EndObject()


def End(builder):
    return OfferEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import random
import timeit

import txaio

txaio.use_twisted()  # noqa

import pytest
from zlmdb import flatbuffers

from cfxdb.xbrmm import ChannelBalance


def fill_channel_balance(channel_balance):
    channel_balance.remaining = random.randint(0, 2**256 - 1)
    channel_balance.inflight = random.randint(0, 2**256 - 1)
    channel_balance.seq = random.randint(0, 2**32 - 1)


@pytest.fixture(scope="function")
def channel_balance():
    _channel_balance = ChannelBalance()
    fill_channel_balance(_channel_balance)
    return _channel_balance


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_channel_balance_roundtrip(channel_balance, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = channel_balance.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 112

    # create python object from bytes (flatbuffes)
    _channel_balance = ChannelBalance.cast(data)

    assert _channel_balance.remaining == channel_balance.remaining
    assert _channel_balance.inflight == channel_balance.inflight
    assert _channel_balance.seq == channel_balance.seq


def test_channel_balance_roundtrip_perf(channel_balance, builder):
    obj = channel_balance.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _channel_balance = ChannelBalance.cast(data)
        if True:
            assert _channel_balance.remaining == channel_balance.remaining
            assert _channel_balance.inflight == channel_balance.inflight
            assert _channel_balance.seq == channel_balance.seq

            scratch["value"] += int(_channel_balance.seq)

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

from cfxdb.common import ConfigurationElement


class WebService(ConfigurationElement):
    """
    Web service:

    * check_web_path_service
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        service_type: Optional[str] = None,
        webcluster_oid: Optional[UUID] = None,
        path: Optional[str] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of this web service.

        :param label: Optional user label of this web service.

        :param description: Optional user description of this web service.

        :param tags: Optional list of user tags on this web service.

        :param path: HTTP URL path of the Web service, eg ``/myapp`` or ``/myapp/dashboard/72``.
        """
        ConfigurationElement.__init__(
            self, oid=oid, label=label, description=description, tags=tags, _unknown=_unknown
        )
        self.service_type = service_type
        self.webcluster_oid = webcluster_oid
        self.path = path

    def other(self, key, default=None):
        return self._unknown.get(key, default)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.service_type != self.service_type:
            return False
        if other.webcluster_oid != self.webcluster_oid:
            return False
        if other.path != self.path:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "webcluster_oid": str(self.webcluster_oid),
                "path": self.path,
                "type": self.service_type,
            }
        )

        if self._unknown:
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["type", "path", "webcluster_oid"]:
                _unknown[k] = data[k]

        webcluster_oid = data.get("webcluster_oid", None)
        assert webcluster_oid is None or (type(webcluster_oid) == str)
        if webcluster_oid:
            webcluster_oid = UUID(webcluster_oid)

        path = data.get("path", None)
        assert path is None or (type(path) == str)

        service_type = data.get("type", None)
        assert service_type is None or (type(service_type) == str)

        obj = WebService(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            service_type=service_type,
            webcluster_oid=webcluster_oid,
            path=path,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import uuid
from datetime import datetime

from cfxdb.gen.user import ActivationToken as ActivationTokenGenFbs
from cfxdb.gen.user.ActivationStatus import ActivationStatus
from cfxdb.gen.user.ActivationType import ActivationType


class ActivationTokenFbs(object):
    """
    CFC user activation token database class for Flatbuffers.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._oid = None
        self._atype = None
        self._status = None
        self._created = None
        self._completed = None
        self._code = None
        self._email = None
        self._pubkey = None

    def copy(self, other):
        self._oid = other.oid
        self._atype = other.atype
        self._status = other.status
        self._created = other.created
        self._completed = other.completed
        self._code = other.code
        self._email = other.email
        self._pubkey = other.pubkey

    @property
    def oid(self):
        if self._oid is None and self._from_fbs:
            s = self._from_fbs.Oid()
            if s:
                self._oid = uuid.UUID(s.decode())

        return self._oid

    @oid.setter
    def oid(self, value):
        assert isinstance(value, uuid.UUID)

        self._oid = value

    @property
    def atype(self):
        if self._atype is None and self._from_fbs:
            self._atype = self._from_fbs.Atype()

        return self._atype

    @atype.setter
    def atype(self, value):
        assert type(value) == int
        assert value in [ActivationType.REGISTRATION, ActivationType.LOGIN]

        self._atype = value

    @property
    def status(self):
        if self._status is None and self._from_fbs:
            self._status = self._from_fbs.Status()

        return self._status

    @status.setter
    def status(self, value):
        assert type(value) == int
        assert value in [ActivationStatus.PENDING, ActivationStatus.ACTIVE, ActivationStatus.EXPIRED]

        self._status = value

    @property
    def created(self):
        if self._created is None and self._from_fbs:
            val = self._from_fbs.Created()
            if val:
                self._created = datetime.utcfromtimestamp(float(val) / 1000000.0)

        return self._created

    @created.setter
    def created(self, value):
        assert isinstance(value, datetime)

        self._created = value

    @property
    def completed(self):
        if self._completed is None and self._from_fbs:
            val = self._from_fbs.Completed()
            if val:
                self._completed = datetime.utcfromtimestamp(float(val) / 1000000.0)

        return self._completed

    @completed.setter
    def completed(self, value):
        assert value is None or isinstance(value, datetime)

        self._completed = value

    @property
    def code(self):
        if self._code is None and self._from_fbs:
            self._code = self._from_fbs.Code().decode()

        return self._code

    @code.setter
    def code(self, value):
        assert type(value) == str

        self._code = value

    @property
    def email(self):
        if self._email is None and self._from_fbs:
            self._email = self._from_fbs.Email().decode()

        return self._email

    @email.setter
    def email(self, value):
        assert type(value) == str

        self._email = value

    @property
    def pubkey(self):
        if self._pubkey is None and self._from_fbs:
            self._pubkey = self._from_fbs.Pubkey().decode()

        return self._pubkey

    @pubkey.setter
    def pubkey(self, value):
        # hex string with 256 bit Ed25519 WAMP-cryptosign public key
        assert type(value) == str
        assert len(value) == 64

        self._pubkey = value

    @staticmethod
    def cast(buf):
        return ActivationTokenFbs(ActivationTokenGenFbs.ActivationToken.GetRootAsActivationToken(buf, 0))

    def build(self, builder):
        oid = self.oid
        if oid:
            oid = builder.CreateString(str(oid))

        code = self.code
        if code:
            code = builder.CreateString(code)

        email = self.email
        if email:
            email = builder.CreateString(email)

        pubkey = self.pubkey
        if pubkey:
            pubkey = builder.CreateString(pubkey)

        ActivationTokenGenFbs.ActivationTokenStart(builder)

        if oid:
            ActivationTokenGenFbs.ActivationTokenAddOid(builder, oid)

        ActivationTokenGenFbs.ActivationTokenAddAtype(builder, self.atype)
        ActivationTokenGenFbs.ActivationTokenAddStatus(builder, self.status)

        if self.created:
            ActivationTokenGenFbs.ActivationTokenAddCreated(builder, int(self.created.timestamp() * 1000000))

        if self.completed:
            ActivationTokenGenFbs.ActivationTokenAddCompleted(builder, int(self.completed.timestamp() * 1000000))

        if code:
            ActivationTokenGenFbs.ActivationTokenAddCode(builder, code)

        if email:
            ActivationTokenGenFbs.ActivationTokenAddEmail(builder, email)

        if pubkey:
            ActivationTokenGenFbs.ActivationTokenAddPubkey(builder, pubkey)

        final = ActivationTokenGenFbs.ActivationTokenEnd(builder)

        return final



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class oid_t(object):
    __slots__ = ["_tab"]

    @classmethod
    def SizeOf(cls):
        return 16

    # oid_t
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # oid_t
    def Msb(self):
        return self._tab.Get(
            flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0)
        )

    # oid_t
    def Lsb(self):
        return self._tab.Get(
            flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8)
        )


def Createoid_t(builder, msb, lsb):
    builder.Prep(8, 16)
    builder.PrependUint64(lsb)
    builder.PrependUint64(msb)
    return builder.Offset()
